#
#  Copyright (2024) Hewlett Packard Enterprise Development LP
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from typing import Optional

from numpy.typing import NDArray

from .types import CamOp, CamVariant, is_float_type


def validate_args(
    variant: CamVariant,
    op: CamOp,
    inputs: NDArray,
    cam: NDArray,
    reduction_values: Optional[NDArray] = None,
) -> None:
    """
    Ensures that the arguments for a CAM operation are sane and compatible
    by inspecting their dtypes and shapes.
    Raises appropriate errors in case a problem is found.
    """

    # dtype checking

    if inputs.dtype != cam.dtype:
        raise TypeError(
            f"inputs and cam dtypes are not equal: ({inputs.dtype} vs {cam.dtype})"
        )

    if op.is_reduction:
        if reduction_values is None:
            raise TypeError("operation is reduction but no reduction values were given")

        if not is_float_type(reduction_values.dtype):
            raise TypeError("reductions only support float data types")

    # shape checking

    if len(inputs.shape) < 2:
        raise ValueError("inputs are one-dimensional")

    if len(cam.shape) < 2:
        raise ValueError("cam is one-dimensional")

    cam_cell_width = variant.cam_cell_width
    if cam.shape[-1] % cam_cell_width != 0:
        raise ValueError(
            "amount of CAM columns is not divisible by CAM cell encoding width: "
            f"{cam_cell_width} (columns found: {cam.shape[-1]})"
        )

    input_cell_width = variant.input_cell_width
    if inputs.shape[-1] % input_cell_width != 0:
        raise ValueError(
            "amount of input columns is not divisible by input cell encoding width: "
            f"{input_cell_width} (columns found: {inputs.shape[-1]})"
        )

    if inputs.shape[-1] // input_cell_width != cam.shape[-1] // cam_cell_width:
        raise ValueError(
            "input and cam column sizes do not match "
            f"a ratio of {input_cell_width}:{cam_cell_width}"
        )

    if op.is_reduction:
        assert reduction_values is not None
        if len(reduction_values.shape) > 1:
            raise ValueError("reduction values must be one-dimensional")
        if len(inputs.shape) > 2 or len(cam.shape) > 2:
            raise ValueError("reduction operations do not support broadcasting")
        if reduction_values.shape != cam.shape[:-1]:
            raise ValueError("CAM shape does not match reduction value shape")

    inputs_outer_shape, cam_outer_shape = inputs.shape[:-2], cam.shape[:-2]

    for i in range(1, min(len(inputs_outer_shape), len(cam_outer_shape)) + 1):
        if inputs_outer_shape[-i] != cam_outer_shape[-i]:
            raise ValueError(
                "the outer shapes of input and cam are incompatible: "
                f"{inputs_outer_shape} vs {cam_outer_shape}"
            )



#
#  Copyright (2024) Hewlett Packard Enterprise Development LP
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
This module contains the `flip_indices` utility function CUDA kernel and public API.
"""

from string import Template

import cupy as cp
from numpy.typing import DTypeLike, NDArray

from ..types import IntDType, NumericDType, dtype_to_ctype, is_float_type
from .helpers import simple_kernel_dimensions

# Kernel notes:
#
# Dimensions:
#    Linear dimensions, i.e. only `threadIdx.x` and `blockIdx.x`.
#
# Operation per thread:
#    Each thread applies a single index in `indices`.

KERNEL_SOURCE_TEMPLATE = Template(
    r"""
extern "C" __global__
void flip_indices(
    $inputs_type *inputs, $indices_type *indices,
    long input_rows, long input_cols, long index_rows, long index_cols
) {
    long thread_id = threadIdx.x + (blockIdx.x * blockDim.x);

    if (thread_id >= input_rows * index_cols) {
        return;
    }

    $indices_type col_index = indices[thread_id % (index_rows * index_cols)];

    // ignore sub-zero indices
    if (col_index < 0) {
        return;
    }

    long row = thread_id / index_cols;
    inputs[row * input_cols + col_index] = 1 - inputs[row * input_cols + col_index];
}
"""
)


def generate_kernel(inputs_dtype: DTypeLike, indices_dtype: DTypeLike):
    """Converts the generic source template into an instance of `cp.RawKernel`"""

    if is_float_type(indices_dtype):
        raise TypeError(f"data type of indices is not an integer ({indices_dtype})")

    code = KERNEL_SOURCE_TEMPLATE.safe_substitute(
        inputs_type=dtype_to_ctype(inputs_dtype),
        indices_type=dtype_to_ctype(indices_dtype),
    )

    return cp.RawKernel(code=code, name="flip_indices")


def flip_indices(inputs: NDArray[NumericDType], indices: NDArray[IntDType]) -> None:
    """
    Flips (`x = 1 - x`) the elements in a two-dimensional input array indexed
    by a two-dimensional array of indices.

    ### Arguments:
    - `inputs` (`cupy.ndarray`): The two-dimensional input array that is
      mutated in-place.
    - `indices` (`numpy.ndarray` or `cupy.ndarray`): The two-dimensional indices.
      All indices less than zero are ignored.

    ### Returns:
    `None`

    ### Notes:
    - This is an in-place operation, so `inputs` is expected to be on the GPU.
    - `indices` can have less rows than `inputs`.
      In that case, `indices` is repeated across the rows of `inputs` such that they
      match.
    """

    if not isinstance(inputs, cp.ndarray):
        raise TypeError("expected inputs to be a cupy array already")

    if len(inputs.shape) < 2:
        raise ValueError("expected inputs to be two-dimensional")

    if len(indices.shape) < 2:
        raise ValueError("expected indices to be two-dimensional")

    index_rows, index_cols = indices.shape[-2], indices.shape[-1]
    input_rows, input_cols = inputs.shape[-2], inputs.shape[-1]

    if input_rows % index_rows != 0:
        raise ValueError(
            f"{index_rows} index rows can not evenly be repeated over {input_rows} input rows"  # noqa: E501
        )

    kernel = generate_kernel(inputs.dtype, indices.dtype)

    dims = simple_kernel_dimensions(
        input_rows * index_cols,  # one thread per index in the (repeated) indices array
        kernel.attributes["max_threads_per_block"],
    )

    indices = cp.asarray(indices.ravel())  # move `indices` to the GPU if needed

    kernel(*dims, (inputs, indices, input_rows, input_cols, index_rows, index_cols))



#
#  Copyright (2024) Hewlett Packard Enterprise Development LP
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
This module contains all CAM related functions.
Below are some general notes that apply to all functions in this module.

### Notes:
- Both the CAM and its inputs are always expected to contain the same data type.
- NumPy arrays passed to functions in this module will automatically be moved to
  the GPU. If you are calling a function multiple times with a fixed subset of
  parameters, consider moving them to the GPU beforehand yourself to avoid
  unnecessary data transfer between the host and the GPU. You can move NumPy arrays
  to the GPU via `cupy.asarray`.
- The arrays returned by functions in this module are always still on the GPU.
  To retrieve them to the host, call `.get()` or `cupy.asnumpy(...)`.

### Broadcasting:
- If both the inputs and the CAM are 2-D, they are matched like a conventional
  set of inputs and CAM.
- If both the inputs and the CAM are N-D, N > 2, they are treated like a stack of
  inputs and CAMs and matching only takes place within the last two dimensions.
- If the inputs are N-D, N >= 2 and the CAMs are M-D, M >= 2, the smaller
  argument is repeatedly applied to the greater one.
- This is analogous to the way that matrix multiplication works in NumPy
  with dimensions larger than 2.
"""

from typing import Optional, TypeVar

import cupy as cp
import numpy as np
from numpy.typing import NDArray

from .kernel import generate_kernel
from .run import run_kernel
from .types import CamOp, CamVariant, FloatDType, NumericDType
from .validation import validate_args

TN = TypeVar("TN", bound=NumericDType)
TF = TypeVar("TF", bound=FloatDType)
UF = TypeVar("UF", bound=FloatDType)


def tcam_match(inputs: NDArray[TN], cam: NDArray[TN]) -> NDArray[np.int8]:
    """
    Performs the matching operation of a TCAM (Ternary CAM) on a given set of inputs and
    TCAMs.

    ### Arguments:
    - `inputs` (`numpy.ndarray` or `cupy.ndarray`): The input columns stacked in rows.
      The two innermost dimensions of `inputs` are of shape `input_rows x columns`.
    - `cam` (`numpy.ndarray` or `cupy.ndarray`): The TCAM matrix itself.
      When using a float dytpe, a "Don't Care" (or "X") is encoded as `numpy.nan`.
      Alternatively, for integer dtypes they are indicated by values < 0 or > 1.
      The two innermost dimensions of `cam` are of shape `cam_rows x columns`.

    ### Returns:
    `cupy.ndarray`: The resulting matrix of matches encoded as `np.int8`s.
    The two innermost dimensions are of shape `input_rows x cam_rows`.
    """
    variant, op, result_dtype = CamVariant.TCAM, CamOp.MATCH, np.int8
    validate_args(variant, op, inputs, cam)
    kernel = generate_kernel(variant, op, inputs.dtype, cam.dtype, result_dtype)
    return run_kernel(kernel, variant, op, inputs, cam, result_dtype)


def tcam_hamming_distance(inputs: NDArray[TN], cam: NDArray[TN]) -> NDArray[np.int64]:
    """
    Determines the hamming distance (number of mismatches per row) in a
    TCAM (Ternary CAM) matching operation for a given set of inputs and TCAMs.

    ### Arguments:
    - `inputs` (`numpy.ndarray` or `cupy.ndarray`): The input columns stacked in rows.
      The two innermost dimensions of `inputs` are of shape `input_rows x columns`.
    - `cam` (`numpy.ndarray` or `cupy.ndarray`): The TCAM matrix itself.
      When using a float dytpe, a "Don't Care" (or "X") is encoded as `numpy.nan`.
      Alternatively, for integer dtypes they are indicated by values < 0 or > 1.
      The two innermost dimensions of `cam` are of shape `cam_rows x columns`.

    ### Returns:
    `cupy.ndarray`: The resulting matrix of mismatch counts encoded as `np.int64`s.
    The two innermost dimensions are of shape `input_rows x cam_rows`.
    """
    variant, op, result_dtype = CamVariant.TCAM, CamOp.COUNT_MISMATCHES, np.int64
    validate_args(variant, op, inputs, cam)
    kernel = generate_kernel(variant, op, inputs.dtype, cam.dtype, result_dtype)
    return run_kernel(kernel, variant, op, inputs, cam, result_dtype)


def tcam_reduce_sum(
    inputs: NDArray[TN], cam: NDArray[TN], *, values: NDArray[TF]
) -> NDArray[TF]:
    """
    First performs the matching operation of a TCAM (Ternary CAM) on a given set
    of inputs and TCAMs, then reduces each row of inputs down to a single number
    by accumulating over a set of values for every matched row.

    Note that reduction operations do not support broadcasting.

    ### Positional Arguments:
    - `inputs` (`numpy.ndarray` or `cupy.ndarray`): The input columns stacked in rows.
      The two innermost dimensions of `inputs` are of shape `input_rows x columns`.
    - `cam` (`numpy.ndarray` or `cupy.ndarray`): The TCAM matrix itself.
      When using a float dytpe, a "Don't Care" (or "X") is encoded as `numpy.nan`.
      Alternatively, for integer dtypes they are indicated by values < 0 or > 1.
      The two innermost dimensions of `cam` are of shape `cam_rows x columns`.

    ### Keyword Arguments:
    - `values` (`numpy.ndarray` or `cupy.ndarray`): The values to reduce.
      The dimensions of `values` must be the same as the dimensions of `cam`,
      except for the omission of the `columns` dimension.

    ### Returns:
    `cupy.ndarray`: The reduction result matrix.
    The innermost dimension is `input_rows`.
    """
    variant, op, result_dtype = CamVariant.TCAM, CamOp.REDUCE_SUM, values.dtype
    validate_args(variant, op, inputs, cam, values)
    kernel = generate_kernel(variant, op, inputs.dtype, cam.dtype, result_dtype)
    return run_kernel(kernel, variant, op, inputs, cam, result_dtype, values)


def add_noise(acam: NDArray, noise: float) -> NDArray:
    """Adds noise to ACAM thresholds based on a normal distribution."""
    xp = cp.get_array_module(acam)
    return acam + xp.random.normal(0, noise, acam.shape).astype(acam.dtype)


def acam_match(
    inputs: NDArray[TN], cam: NDArray[TN], *, noise: Optional[float] = None
) -> NDArray[np.int8]:
    """
    Performs the matching operation of a ACAM (Analog CAM) on a given set of inputs and
    ACAMs.

    ### Positional Arguments:
    - `inputs` (`numpy.ndarray` or `cupy.ndarray`): The input columns stacked in rows.
      The two innermost dimensions of `inputs` are of shape `input_rows x columns`.
    - `cam` (`numpy.ndarray` or `cupy.ndarray`): The ACAM matrix itself.
      The lower and upper thresholds are encoded side-by-side within the column.
      This means that your CAM columns should be twice as wide as your input columns.
      "Don't Care" (or "X") thresholds are encoded as `numpy.nan` values when using
      float types and as negative integers when using integer types.
      The two innermost dimensions of `cam` are of shape `cam_rows x (columns * 2)`.

    ### Keyword Arguments:
    - `noise` (`float` [optional]): Standard deviation for a normal distribution
      `N(0, noise)` that is randomly sampled from and added to the ACAM thresholds
      to simulate analog inaccuracies before performing the operation. Noise is only
      added if this argument is defined.

    ### Returns:
    `cupy.ndarray`: The resulting matrix of matches encoded as `np.int8`s.
    The two innermost dimensions are of shape `input_rows x cam_rows`.

    ### Notes:
    - The arguments for ACAM operations must contain floating point data types.
    """
    variant, op, result_dtype = CamVariant.ACAM, CamOp.MATCH, np.int8
    validate_args(variant, op, inputs, cam)
    kernel = generate_kernel(variant, op, inputs.dtype, cam.dtype, result_dtype)
    if noise is not None:
        cam = add_noise(cam, noise)
    return run_kernel(kernel, variant, op, inputs, cam, result_dtype)


def acam_count_mismatches(
    inputs: NDArray[TN], cam: NDArray[TN], *, noise: Optional[float] = None
) -> NDArray[np.int64]:
    """
    Determines the number of mismatches per row in a ACAM (Analog CAM) matching
    operation for a given set of inputs and ACAMs.

    ### Positional Arguments:
    - `inputs` (`numpy.ndarray` or `cupy.ndarray`): The input columns stacked in rows.
      The two innermost dimensions of `inputs` are of shape `input_rows x columns`.
    - `cam` (`numpy.ndarray` or `cupy.ndarray`): The ACAM matrix itself.
      The lower and upper thresholds are encoded side-by-side within the column.
      This means that your CAM columns should be twice as wide as your input columns.
      "Don't Care" (or "X") thresholds are encoded as `numpy.nan` values when using
      float types and as negative integers when using integer types.
      The two innermost dimensions of `cam` are of shape `cam_rows x (columns * 2)`.

    ### Keyword Arguments:
    - `noise` (`float` [optional]): Standard deviation for a normal distribution
      `N(0, noise)` that is randomly sampled from and added to the ACAM thresholds
      to simulate analog inaccuracies before performing the operation. Noise is only
      added if this argument is defined.

    ### Returns:
    `cupy.ndarray`: The resulting matrix of mismatch counts encoded as `np.int64`s.
    The two innermost dimensions are of shape `input_rows x cam_rows`.

    ### Notes:
    - The arguments for ACAM operations must contain floating point data types.
    """
    variant, op, result_dtype = CamVariant.ACAM, CamOp.COUNT_MISMATCHES, np.int64
    validate_args(variant, op, inputs, cam)
    kernel = generate_kernel(variant, op, inputs.dtype, cam.dtype, result_dtype)
    if noise is not None:
        cam = add_noise(cam, noise)
    return run_kernel(kernel, variant, op, inputs, cam, result_dtype)


def acam_reduce_sum(
    inputs: NDArray[TN],
    cam: NDArray[TN],
    *,
    values: NDArray[UF],
    noise: Optional[float] = None,
) -> NDArray[UF]:
    """
    First performs the matching operation of a ACAM (Analog CAM) on a given set
    of inputs and ACAMs, then reduces each row of inputs down to a single number
    by accumulating over a set of values for every matched row.

    Note that reduction operations do not support broadcasting.

    ### Positional Arguments:
    - `inputs` (`numpy.ndarray` or `cupy.ndarray`): The input columns stacked in rows.
      The two innermost dimensions of `inputs` are of shape `input_rows x columns`.
    - `cam` (`numpy.ndarray` or `cupy.ndarray`): The ACAM matrix itself.
      The lower and upper thresholds are encoded side-by-side within the column.
      This means that your CAM columns should be twice as wide as your input columns.
      "Don't Care" (or "X") thresholds are encoded as `numpy.nan` values when using
      float types and as negative integers when using integer types.
      The two innermost dimensions of `cam` are of shape `cam_rows x (columns * 2)`.

    ### Keyword Arguments:
    - `values` (`numpy.ndarray` or `cupy.ndarray`): The values to reduce.
      The dimensions of `values` must be the same as the dimensions of `cam`,
      except for the omission of the `columns` dimension.
    - `noise` (`float` [optional]): Standard deviation for a normal distribution
      `N(0, noise)` that is randomly sampled from and added to the ACAM thresholds
      to simulate analog inaccuracies before performing the operation. Noise is only
      added if this argument is defined.

    ### Returns:
    `cupy.ndarray`: The reduction result matrix.
    The innermost dimension is `input_rows`.

    ### Notes:
    - The arguments for ACAM operations must contain floating point data types.
    """
    variant, op, result_dtype = CamVariant.ACAM, CamOp.REDUCE_SUM, values.dtype
    validate_args(variant, op, inputs, cam, values)
    kernel = generate_kernel(variant, op, inputs.dtype, cam.dtype, result_dtype)
    if noise is not None:
        cam = add_noise(cam, noise)
    return run_kernel(kernel, variant, op, inputs, cam, result_dtype, values)


def acam_double_match(
    inputs: NDArray[TN], cam: NDArray[TN], *, noise: Optional[float] = None
) -> NDArray[np.int8]:
    """
    Performs the matching operation of a ACAM (Analog CAM) with double precision on a given set of inputs and
    ACAMs.

    ### Positional Arguments:
    - `inputs` (`numpy.ndarray` or `cupy.ndarray`): The input columns stacked in rows.
      The two innermost dimensions of `inputs` are of shape `input_rows x columns`.
    - `cam` (`numpy.ndarray` or `cupy.ndarray`): The ACAM matrix itself.
      The lower and upper thresholds are encoded side-by-side within the column.
      This means that your CAM columns should be twice as wide as your input columns.
      "Don't Care" (or "X") thresholds are encoded as `numpy.nan` values when using
      float types and as negative integers when using integer types.
      The two innermost dimensions of `cam` are of shape `cam_rows x (columns * 2)`.

    ### Keyword Arguments:
    - `noise` (`float` [optional]): Standard deviation for a normal distribution
      `N(0, noise)` that is randomly sampled from and added to the ACAM thresholds
      to simulate analog inaccuracies before performing the operation. Noise is only
      added if this argument is defined.

    ### Returns:
    `cupy.ndarray`: The resulting matrix of matches encoded as `np.int8`s.
    The two innermost dimensions are of shape `input_rows x cam_rows`.

    ### Notes:
    - The arguments for ACAM operations must contain floating point data types.
    """
    variant, op, result_dtype = CamVariant.ACAM_DOUBLE, CamOp.MATCH, np.int8
    validate_args(variant, op, inputs, cam)
    kernel = generate_kernel(variant, op, inputs.dtype, cam.dtype, result_dtype)
    if noise is not None:
        cam = add_noise(cam, noise)
    return run_kernel(kernel, variant, op, inputs, cam, result_dtype)


def acam_double_count_mismatches(
    inputs: NDArray[TN], cam: NDArray[TN], *, noise: Optional[float] = None
) -> NDArray[np.int64]:
    """
    Determines the number of mismatches per row in a ACAM (Analog CAM) matching
    operation for a given set of inputs and ACAMs.

    ### Positional Arguments:
    - `inputs` (`numpy.ndarray` or `cupy.ndarray`): The input columns stacked in rows.
      The two innermost dimensions of `inputs` are of shape `input_rows x columns`.
    - `cam` (`numpy.ndarray` or `cupy.ndarray`): The ACAM matrix itself.
      The lower and upper thresholds are encoded side-by-side within the column.
      This means that your CAM columns should be twice as wide as your input columns.
      "Don't Care" (or "X") thresholds are encoded as `numpy.nan` values when using
      float types and as negative integers when using integer types.
      The two innermost dimensions of `cam` are of shape `cam_rows x (columns * 2)`.

    ### Keyword Arguments:
    - `noise` (`float` [optional]): Standard deviation for a normal distribution
      `N(0, noise)` that is randomly sampled from and added to the ACAM thresholds
      to simulate analog inaccuracies before performing the operation. Noise is only
      added if this argument is defined.

    ### Returns:
    `cupy.ndarray`: The resulting matrix of mismatch counts encoded as `np.int64`s.
    The two innermost dimensions are of shape `input_rows x cam_rows`.

    ### Notes:
    - The arguments for ACAM operations must contain floating point data types.
    """
    variant, op, result_dtype = CamVariant.ACAM_DOUBLE, CamOp.COUNT_MISMATCHES, np.int64
    validate_args(variant, op, inputs, cam)
    kernel = generate_kernel(variant, op, inputs.dtype, cam.dtype, result_dtype)
    if noise is not None:
        cam = add_noise(cam, noise)
    return run_kernel(kernel, variant, op, inputs, cam, result_dtype)


def acam_double_reduce_sum(
    inputs: NDArray[TN],
    cam: NDArray[TN],
    *,
    values: NDArray[UF],
    noise: Optional[float] = None,
) -> NDArray[UF]:
    """
    First performs the matching operation of a ACAM (Analog CAM) on a given set
    of inputs and ACAMs, then reduces each row of inputs down to a single number
    by accumulating over a set of values for every matched row.

    Note that reduction operations do not support broadcasting.

    ### Positional Arguments:
    - `inputs` (`numpy.ndarray` or `cupy.ndarray`): The input columns stacked in rows.
      The two innermost dimensions of `inputs` are of shape `input_rows x columns`.
    - `cam` (`numpy.ndarray` or `cupy.ndarray`): The ACAM matrix itself.
      The lower and upper thresholds are encoded side-by-side within the column.
      This means that your CAM columns should be twice as wide as your input columns.
      "Don't Care" (or "X") thresholds are encoded as `numpy.nan` values when using
      float types and as negative integers when using integer types.
      The two innermost dimensions of `cam` are of shape `cam_rows x (columns * 2)`.

    ### Keyword Arguments:
    - `values` (`numpy.ndarray` or `cupy.ndarray`): The values to reduce.
      The dimensions of `values` must be the same as the dimensions of `cam`,
      except for the omission of the `columns` dimension.
    - `noise` (`float` [optional]): Standard deviation for a normal distribution
      `N(0, noise)` that is randomly sampled from and added to the ACAM thresholds
      to simulate analog inaccuracies before performing the operation. Noise is only
      added if this argument is defined.

    ### Returns:
    `cupy.ndarray`: The reduction result matrix.
    The innermost dimension is `input_rows`.

    ### Notes:
    - The arguments for ACAM operations must contain floating point data types.
    """
    variant, op, result_dtype = CamVariant.ACAM_DOUBLE, CamOp.REDUCE_SUM, values.dtype
    validate_args(variant, op, inputs, cam, values)
    kernel = generate_kernel(variant, op, inputs.dtype, cam.dtype, result_dtype)
    if noise is not None:
        cam = add_noise(cam, noise)
    return run_kernel(kernel, variant, op, inputs, cam, result_dtype, values)



#
#  Copyright (2024) Hewlett Packard Enterprise Development LP
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Python APIs to simulate various CAMs (Content Addressable Memory) on GPUs at scale.
"""

import importlib.metadata

from .cam import (
    acam_count_mismatches,
    acam_double_count_mismatches,
    acam_double_match,
    acam_double_reduce_sum,
    acam_match,
    acam_reduce_sum,
    tcam_hamming_distance,
    tcam_match,
    tcam_reduce_sum,
)
from .util import flip_indices

__all__ = [
    "acam_count_mismatches",
    "acam_double_count_mismatches",
    "acam_double_match",
    "acam_double_reduce_sum",
    "acam_match",
    "acam_reduce_sum",
    "flip_indices",
    "tcam_hamming_distance",
    "tcam_match",
    "tcam_reduce_sum",
]

# https://github.com/python-poetry/poetry/issues/144#issuecomment-1488038660
__version__ = importlib.metadata.version("campie")
